#!/usr/bin/env ruby

=begin
= class DBIAccessor

== Synopsis

   See the sample program.

== Description

(({DBIAccessor})) is a glue class between the DBBackedClass object
space and the database itself.  While rarely used directly, it provides
the engine for DBBackedClasses to reconstitute instances from data in
the database, create new database records from new instances, and
delete existing instances from the database.

== Class Methods

--- DBIAccessor.new(dbh, query_strings)
    Constructs a new (({DBIAccessor})) instance, talking to the database
    attached to the (({DBI})) database handle (({dbh})) with the query
    strings being a hash of tags to parameterized SQL query strings.  This
    allows (({DBIAccessor})) to be database independent, in the true
    spirit of (({DBI})).  Want to backend with a different DB?  Just drop
    in a new set of query strings and pass a database handle to that new
    database.

== Instance Methods

You will probably never use any of these directly.  Rather they get
called by instances of (({DBBackedClass})).

--- DBIAccessor#selexec2(stmt, *params)

    Execute the statement stmt with the parameters params (an arbitrary
    number of parameters to the method).  Results are return as an array
    of rows.  Each row is a hash of keys (field names) to values.

--- DBIAccessor#selexec(query, *params)

    This is similar to ((<DBIAccessor#selexec2>)) only query is the
    shorthand tag for one of the queries in the query string hash.  The
    appropriate query is looked up and then it and the parameters are
    passed to (({selexec2})).

--- DBIAccessor#listByType(klass, modifier=nil)

    List all instances of type (({klass})) where (({klass})) is an
    instance of (({Class})).  Optionally (({modifier})) may be a non-null
    string modifying the query. An array of results is returned.

--- DBIAccessor#lookupByTypeAndID(klass, record_id)

    Lookup a particular instance/record from the table (({klass})) by
    record ID where (({klass})) is an instance of (({Class})).  Return the
    result as an array of rows where each row is a hash of key (field name)
    => value.

--- DBIAccessor#lookupByAttributes(object)

    Given an object (({object})), we try to find all rows that look like it
    (by looking at the attributes that are actually defined for
    (({object}))).  Return the results as an array of rows where each row
    is a hash of key (field name) => value.

--- DBIAccessor#lookupLastInsert(object)

    Lookup the last record insert into the table which stores instances of
    the same class as (({object})).  Used internally by DBIAccessor.

--- DBIAccessor#insert(object)

    Insert (({object})) into the table.  Fails if an instance with the same
    record ID already exists.

--- DBIAccessor#update(object)

    Update (({object})) in the database.

--- DBIAccessor#delete(object)

    Delete (({object})) from the database.
=end
class DBIAccessor
	attr_accessor :dbh
	attr_accessor :debug

	def object_to_table(object)
  		klass = object.type == Class ? object : object.class
		# Chop off the module name if there is one.
		arr = klass.to_s.split('::')
		t = arr[1] || arr[0]
		return eval (':'+t)
  	end

	def initialize(dbh,query_strings)
		@dbh = dbh
		@QUERY_STRINGS = query_strings
	end

	def selexec2(stmt,*params)
		if(debug)
  			$stderr.puts stmt
  			$stderr.puts params.inspect
			actquery=stmt.clone
			p=params.clone
			actquery.gsub!('\?') { |sub| p.shift }
			$stderr.puts "QUERY: #{actquery}\n"
		end
		st = @dbh.prepare(stmt)
		st.execute(*params)
		rows = []
		st.fetch_all.each { |row| rows << row.to_h }
		st.finish
		if(debug)
			rows.each do |row|
				row.each{|k,v| $stderr.puts "#{k} => #{v}" }
			end
		end
		return rows
	end

	def selexec(query,*params)
		return selexec2(@QUERY_STRINGS[query],*params)
	end

	def fetchRandom(klass,modifier=nil)
		stmt = modifier ? @QUERY_STRINGS['lookupByRandomAndModifier'].gsub('%modifier',modifier) : @QUERY_STRINGS['lookupByRandom']
		return selexec2(stmt,object_to_table(klass))[0]
	end

	def listByType(klass,modifier=nil)
		stmt = modifier ? @QUERY_STRINGS['listByTypeAndModifier'].gsub('%modifier',modifier) : @QUERY_STRINGS['listByType']
		return selexec2(stmt,object_to_table(klass))
	end

	def countByType(klass,modifier=nil)
		stmt = modifier ? @QUERY_STRINGS['countByTypeAndModifier'].gsub('%modifier',modifier) : @QUERY_STRINGS['countByType']
		# Grab first row, flatten hash to array, get the first
		# pair, and the value of that.  Got it? :)
		return selexec2(stmt,object_to_table(klass))[0].to_a[0][1].to_i
	end

	def lookupByTypeAndID(klass,record_id)
		return selexec('lookupByTypeAndID',object_to_table(klass),record_id)[0]
	end

	def lookupByAttributes(object)
		tests = ''
		args = []
		object.attributes.each do |k,v|
			arg = object.hashtable[v[0]]
			next if(v[0] == object.primary_key_name && arg == nil)
			tests += "#{v[0]}"+(arg ? '=' : ' IS ')+"? AND "
			args << arg
		end
		object.refs.each do |k,v|
			arg = object.hashtable[v[0]]
			tests += "#{v[0]}"+(arg ? '=' : ' IS ')+"? AND "
			args << arg
		end
		tests.chomp!('AND ')
		table = object_to_table(object).to_s
		stmt =@QUERY_STRINGS['lookupByAttributes'].
					 gsub('%table',table).
			                 gsub('%where',tests)
		$stderr.puts stmt if @debug
		$stderr.puts args.inspect if @debug
		st = @dbh.prepare(stmt)
		st.execute(*args)
		rows = []
		st.fetch_all.each { |row| rows << row.to_h }
		st.finish
		if(debug)
			rows.each {|row|
				row.each{|k,v| $stderr.puts "#{k} => #{v}" }
			}
		end
		return rows[0]
	end

	def lookupLastInsert(object)
		return selexec('lookupLastInsert',object.class)[0]
	end

	def insert(object)
		argnames = ''
		valholders = ''
		args = []
		object.hashtable.each do |k,v|
			if(v != nil)
				argnames += "#{k},"
				valholders += '?,'
				args << v
			end
		end
		argnames.chomp!(',')
		valholders.chomp!(',')
		table = object_to_table(object).to_s
		stmt =@QUERY_STRINGS['insert'].gsub('%table',table).
		                               gsub('%argnames',argnames).
			                       gsub('%valholders',valholders)
  		$stderr.puts stmt if @debug
		$stderr.puts args.inspect if @debug
		st = @dbh.prepare(stmt)
		st.execute(*args)
		st.finish

		if(object.hashtable[object.primary_key_name]==nil)
			return lookupLastInsert(object)
  		else
			return object.hashtable
		end
 	end

	def update(object)
		argnames = ''
		valholders = ''
		args = []
		object.hashtable.each do |k,v|
			if(v != nil)
				argnames += "#{k},"
				valholders += '?,'
				args << v
			end
		end
		argnames.chomp!(',')
		valholders.chomp!(',')
		table = object_to_table(object).to_s
		stmt =@QUERY_STRINGS['update'].gsub('%table',table).
		                               gsub('%argnames',argnames).
			                       gsub('%valholders',valholders)
  		$stderr.puts stmt if @debug
		$stderr.puts args.inspect if @debug
		st = @dbh.prepare(stmt)
		st.execute(*args)
		st.finish

		return object.hashtable
 	end

	def delete(object)
		stmt =@QUERY_STRINGS['delete'].
			gsub('%table',object_to_table(object).to_s).
			gsub('%where',object.primary_key_name+'=?')
		$stderr.puts stmt if @debug
		st = @dbh.prepare(stmt)
		st.execute(object.record_id)
		st.finish
	end
end
