#!/usr/bin/env ruby

require 'date';
=begin
= module IDEquality

== Description

(({IDEquality})) simply provides a convenient mechanism for comparing two
(({DBBackedClass})) instances.  If they are of the same class and they
have the same record ID, then they should be the same instance.

=end
module IDEquality
	def ==(other)
  		# Because otherwise we end up doing nil.record_id
  		return false if(nil == other)
		return self.class == other.class &&
		       self.record_id==other.record_id
	end
end

=begin
= class DBBackedArray

== Description

(({DBBackedArray})) isn't DB backed necessarily, but contains a collection
of objects that are meant to be.  It's really an array that may be
declared to be writeable or readonly.  If it's readonly, then attempts to
insert elements will raise an exception.  This is used by
(({DBBackedClass})) for (({hashed_backref})).

== Class Methods

--- DBBackedArray.new(writeable, &block)

    Construct a new (({DBBackedArray})) which is writeable as dictated by
    the writeable flag.  If it is writeable, then block will be called
    with the new element upon insertion.  This is typically used so that
    when a new element is inserted, it is flushed to the database.

== Instance Methods

--- DBBackedArray#<<(val)

    Insert val into the (({DBBackedArray})) and call the block passed to
    the constructor with (({val})) as a parameter.  This will raise an
    exception if the (({DBBackedArray})) is readonly.
=end
class DBBackedArray < Array
	attr_accessor :writeable
	def initialize(writeable,&block)
		@writeable = writeable
		@block = block
	end

	def <<(val)
  		raise "Array read-only" unless @writeable
  		@block.call(val)
		super(val)
	end
end

=begin
= class DBBackedClass

== Synopsis

See (({sample.rb})) in the (({sample})) directory.

== Description

This is the base class for all database backed classes.  It contains some
useful methods for instances as well as some class methods which allow for
extension of subclasses based on a simple run-time declarations (similar
to the (({attr})) family.  It's a bit messy as I didn't really know what I
was doing at the time and I should probably clean things up now that I know
the proper way to do things.  Nevertheless it works pretty well though
each subclass takes up a tiny bit of extra memory in a class-variable for
the declared attributes.

You will probably never use this class's methods directly.

== Class Methods

--- DBBackedClass.new(accessor, record_id = -1, hash = nil)

    Construct a new instance of the (({DBBackedClass})) with
    (({DBIAccessor})) (({accessor})) whose record ID is (({record_id}))
    which defaults to -1 indicating a new instance and with properties
    contained in (({hash})) a hash of name to values similar to what one
    would get back from (({DBIAccessor#selexec})).

    Instantiating this class directly isn't very useful.  You want
    subclasses.

--- DBBackedClass.register(klass)

    This registers (({klass})).  Basically it sets up some hashes to
    default values for the event where attributes weren't declared.
    Invoked by the constructor.


--- DBBackedClass.attributes(klass)

    Return all (({hashed_attr}))s defined for (({klass})).  Used
    internally.

--- DBBackedClass.refs(klass)

    Return all (({hashed_ref}))s defined for (({klass})).  Used
    internally.

--- DBBackedClass.backrefs(klass)

    Return all (({hashed_backref}))s defined for (({klass})).  Used
    internally.

--- DBBackedClass.hashed_attr(symbol, key, datatype, writeable)

    Define an attribute for a subclass of (({DBBackedClass})).  This is
    used similar to the popular (({attr_reader})), (({attr_accessor})),
    (({attr})), etc.

    Example usage:
      class DumbClass < DBBackedClass
        hashed_attr :record_id, 'record_id', 'i', false
      end

      foo = DumbClass.new(accessor,23)
      foo.record_id == 23

      bar = DumbClass.new(accessor)
      bar.record_id == -1

    This says that DumbClass has a single hashed attribute accessed with
    the (({record_id})) method which will return the value of the
    (({record_id})) field in this particular record, casting it to an
    (({Integer})).  It is not writeable.

--- DBBackedClass.hashed_ref(symbol, key, klass, writeable)

    This is similar to ((<DBBackedClass.hashed_attr>)) only it's used to
    define a method that will return an instance of another
    (({DBBackedClass})) subclass referenced by a foreign key in the
    database.

    Example:
      class DumbClass < DBBackedClass
        hashed_attr :record_id, 'record_id', 'i', false
        hashed_ref :person, 'people_id', Person, true
      end

      foo = DumbClass.new(accessor,23)
      foo.record_id == 23

      foo.person.type == Person

    This says we want a method (({person})) which will lookup a
    (({Person})) from the (({Person})) table whose (({record_id})) field
    matches the value of the (({people_id})) field in this record (a many
    to one or one to one mapping).
    We can change the value of (({person})) with (({foo.person = newguy}))
    where (({newguy})) is some new (({Person})).  Note that this does
    ((*not*)) delete the old (({Person})) as other records may still be
    referencing it.  It only changes this record to reference
    (({newguy})).

--- DBBackedClass.hashed_backref(symbol, key, klass, writeable)

    This is similar to ((<DBBackedClass.hashed_ref>)) only it's for when
    another table has records which refer back to this record (a one to
    many or one to one mapping).  So for example, if our trusty
    (({DumbClass})) had phone numbers associated with it now, we might
    declare it like:

      class DumbClass < DBBackedClass
        hashed_attr :record_id, 'record_id', 'i', false
        hashed_ref :person, 'people_id', Person, true
        hashed_backref :phones, 'dumb_id', Phone, false
      end

      foo = DumbClass.new(accessor,23)

    This says that in the (({Phone})) table, there are records whose
    (({dumb_id})) field may match our record ID.  When we call
    (({foo.phones})) we'll get back an array of (({Phone})) instances
    which point to this (({DumbClass})) instance.  The array is ((*not*))
    writeable, so we can't add new phones.  I guess we can delete them,
    but that's a bug and won't really affect the persistent data.

--- DBBackedClass.primary_key_name

    This returns the primary key name for this class in the database.  In
    theory, your databases could then have a primary key named other than
    'record_id' though I'm pretty sloppy about it and in practice it's
    probably not really a safe thing to rely.  In the future I'll be
    better, I promise.

--- DBBackedClass.count(accessor, modifier=nil)

    Count the number of instances this class has in the database.  Where
    may be an optional WHERE clause.

--- DBBackedClass.fetchAll(accessor,modifier=nil)

    Use (({accessor})) to fetch all instances of this class from the
    class's table.  Optionally, you can specify a where class like
    (({WHERE foo='bar'})) which will limit the result set.

    Actually used a lot like

      DumbClass.fetchAll(accessor).each { |dumb_inst| blah_blah(dumb_inst) }

== Instance Methods

--- DBBackedClass#attributes

    This simply returns a hash of attributes for this class.  It's used
    internally on occassion.

--- DBBackedClass#refs

    This simply returns a hash of references for this class.  It's used
    internally on occassion.

--- DBBackedClass#backrefs

    This simply returns a hash of back references for this class.  It's
    used internally on occassion.

--- DBBackedClass#primary_key_name

    See ((<DBBackedClass.primary_key_name>))

--- DBBackedClass#sanitize

    Occassionally it may be necessary to go through and sanitize the data
    contained in the object's hash before writing out to the database.
    This is particularly true when the object has a boolean field and
    we've set it manually (through the hash) to be 'true' or 'false'
    whereas many databases use 1 or 0.  This will force an update in the
    hash table.  It's really only necessary if code external to the class
    directly manipulates the hash table.

--- DBBackedClass#parseBool

    A method to parse a boolean value from a string which may be 'true',
    'false', '0', '1', or nil (false).

--- DBBackedClass#parseDate(str)

    Parse a date from a string.

--- DBBackedClass#parseTime(str)

    Parse a time from a string.

--- DBBackedClass#to_s

    Format the contents of this object as a String.

--- DBBackedClass#flush

    Flush this to the backing database.  Any references and backrefs will
    also be flushed if they've been loaded.

--- DBBackedClass#fetchRef(refname)

    Referenced objects aren't fetched automatically, only upon being
    referenced through the accessor method.  At this point, this method is
    called and the object is fectched.

--- DBBackedClass#fetchBackrefs(refname)

    This is similar to ((<DBBackedClass#fetchRef>)) but for backrefs.

=end
class DBBackedClass
	include IDEquality
	attr_reader :hashtable, :accessor
	attr_accessor :dirty, :insertme

	# Totally messed up.  This is a master REFS/ATTRIBUTES table for ALL
	# DBBackedClass derivatives.  Individual classes are hash keys.  The
	# values are hashes where the symbols are keys.  So it's like:
	#
	# @@REFS[klass][symbol] = [ db_field_name, Class, writeable? ]
	# @@ATTRIBUTES[klass][symbol] = [ db_field_name, type, writeable? ]
	@@REFS = {}
	@@BACKREFS = {}
	@@ATTRIBUTES = {}
	@@CACHE = {}

	def DBBackedClass.version() return '$Id: db-support.rb,v 1.15 2003/02/24 17:39:09 tschroeder Exp $' end
	
	def sanitize
		attributes.each do |sym,attr|
			name = attr[0]
			if(attr[2])
				method(name+'=').call(method(name).call)
			else
				hashtable[attr[0]]=method(attr[0]).call.to_s
			end
		end
	end

	def parseBool(str)
  		return false if str == nil
  		return true if str.to_i==1 || str.to_s.downcase=='true'
  		return false if str.to_i==0 || str.to_s.downcase=='false'
		return false
	end

	def parseTime(str)
		return nil if str == nil
		if match = Regexp.new('(\d{4})'+
		                      '(\d{2})'+
				      '(\d{2})'+
				      '(\d{2})'+
				      '(\d{2})'+
				      '(\d{2})').match(str)
			yr = match[1].to_i
			mo = match[2].to_i
			dy = match[3].to_i
			hr = match[4].to_i
			mn = match[5].to_i
			sc = match[6].to_i
		end
		return Time.local(yr,mo,dy,hr,mn,sc)
	end

	def parseDate(str)
		return nil if str == nil
		d = m = y = 0
		if match = Regexp.new('(\d{4,4})-0?(\d+)-0?(\d+)').match(str)
			d = match[3].to_i
			m = match[2].to_i
			y = match[1].to_i
		end
		return nil if(d == 0 || m == 0 || y == 0)
		return Date.new(y,m,d)
	end

	def DBBackedClass.register(klass)
  		@@REFS[klass] = {} if @@REFS[klass] == nil
  		@@BACKREFS[klass] = {} if @@BACKREFS[klass] == nil
  		@@ATTRIBUTES[klass] = {} if @@ATTRIBUTES[klass] == nil
	end

	def DBBackedClass.refs(klass) return @@REFS[klass] end
	def DBBackedClass.backrefs(klass) return @@BACKREFS[klass] end
	def DBBackedClass.attributes(klass) return @@ATTRIBUTES[klass] end

	def DBBackedClass.hashed_cached
		self.module_eval '@@cached=true'
	end

	def DBBackedClass.hashed_ref(symbol, key, klass, writeable)
  		DBBackedClass.register(self)
  		@@REFS[self][symbol] = [ key, klass, writeable ]
		reader = <<-EOS
def #{symbol}
fetchRef(:#{symbol}) if nil == @#{symbol}
@#{symbol}=nil if @#{symbol} && @#{symbol}.record_id == -2
return @#{symbol}
end
EOS
		self.module_eval reader
	
		writer = <<-EOS
def #{symbol}=(val)
@dirty = true
@hashtable['#{key}'] = val.record_id.to_s
return @#{symbol}=val
end
EOS
		self.module_eval writer if writeable
	end

	def DBBackedClass.hashed_backref(symbol, key, klass, writeable)
  		DBBackedClass.register(self)
  		@@BACKREFS[self][symbol] = [ key, klass, writeable ]
		reader = <<-EOS
def #{symbol}
fetchBackrefs(:#{symbol}) if(@#{symbol} == nil)
return @#{symbol}
end
EOS
		self.module_eval reader
	end

	def DBBackedClass.hashed_attr(symbol,key,datatype,writeable)
  		DBBackedClass.register(self)
  		@@ATTRIBUTES[self][symbol] = [ key, datatype, writeable ]
		val = "hashtable['#{key}']"
		case datatype
			when nil
			when 'b'
				val = "parseBool(#{val})"
			when 'd'
				val = "parseDate(#{val})"
			when 't'
				val = "parseTime(#{val})"
			else
				val = val + '.to_'+datatype
		end
		reader = <<-EOS
def #{symbol}()
  return hashtable['#{key}'] ? #{val} : nil
end
EOS
		self.module_eval reader
		if (writeable)
		writer = <<-EOS
def #{symbol}=(val)
@dirty = true
if('#{datatype}' == 'b')
	val = val==true ? 1 : 0
elsif('#{datatype}' == 't')
	mon = ( val.month < 10 ? '0' : '' )+ val.mon.to_s
	day = ( val.day < 10 ? '0' : '' )+ val.day.to_s
	hour = ( val.hour < 10 ? '0' : '' )+ val.hour.to_s
	min = ( val.min < 10 ? '0' : '' )+ val.min.to_s
	sec = ( val.sec < 10 ? '0' : '' )+ val.sec.to_s

	val = val.year.to_s + mon + day + hour + min + sec
end
return hashtable['#{key}'] = val.to_s
end
EOS
		self.module_eval writer
		end
	end
	def DBBackedClass.primary_key_name() 'record_id' end

	def refs() DBBackedClass.refs(self.class) end
	def backrefs() DBBackedClass.backrefs(self.class) end
	def attributes() DBBackedClass.attributes(self.class) end
	def primary_key_name() DBBackedClass.primary_key_name end

	def DBBackedClass.random(accessor, modifier=nil)
		result = accessor.fetchRandom(self, modifier)
		return self.new(accessor,result['record_id'].to_i,result)
	end

	# We don't check cached status here because it's too difficult to
	# keep track of whether cached items are from the modifier clause or
	# not, so don't worry about it.  fetchAll always implies going to
	# the database.
	def DBBackedClass.fetchAll(accessor, modifier=nil)
  		results = []
		accessor.listByType(self, modifier).each do |row|
			results << self.new(accessor,row['record_id'].to_i,row)
		end
		return results
	end

	def DBBackedClass.cached()
		return self.module_eval 'defined?(@@cached) ? @@cached : false'
	end

	def DBBackedClass.count(accessor, modifier=nil)
		return accessor.countByType(self,modifier)
	end

	# This allows us to intercept requests for cached classes!
	def DBBackedClass.new(accessor, record_id = -1, hash = nil)
		if(accessor != nil && record_id.to_i > -1 && hash == nil &&
		   self.cached)
		   	return cache_lookup(self,accessor,record_id) if
		   	       cache_lookup(self,accessor,record_id)
		end
		return super(accessor,record_id,hash)
	end

	# Call with just the accessor to create a new one that can be
	# flushed back to the table later.  Call with the accessor and an
	# id to create one that will be fetched from the DB.
	# Call with the accessor, the id, and a hash to create one from
	# data already fetched.
	# 
	# When id == -1 it means we created this by hand.
	# when id == -2 it means that a lookup failed.
	def initialize(accessor, record_id = -1, hash = nil)
  		DBBackedClass.register(self.class)
		@hashtable = hash
		@record_id = record_id
		@accessor = accessor
		@insertme = @dirty = (record_id == -1) ? true : false
		if(@record_id != -1 && nil == @hashtable)
			@hashtable=@accessor.lookupByTypeAndID(self.type,@record_id)
  			# Failed the lookup, let's forge ahead.
  			if (nil == @hashtable)
  				@record_id = -2
  				@hashtable = { 'record_id' => @record_id.to_s }
				@insertme = @dirty = false
			end
		elsif (nil == @hashtable)
			@hashtable = { 'record_id' => @record_id.to_s }
		else
			@hashtable['record_id'] = @record_id.to_s
		end
		cache if self.class.cached
	end

	def to_s
		attrs = []
		@hashtable.each do |k,v|
			attrs << "#{k} => #{v}"
		end
		return '<'+self.class.to_s+': ' + attrs.join(', ') + '>'

	end

	def DBBackedClass.cache_lookup(klass,accessor,record_id)
		rid = record_id ? record_id.to_i : -1
		return nil unless klass && accessor && rid > -1
		return @@CACHE[accessor][klass][rid] if
			@@CACHE[accessor] &&
			@@CACHE[accessor][klass]
		return nil
	end

	def cache
		return if (self.accessor == nil)
		@@CACHE[self.accessor] = {} unless @@CACHE[self.accessor]
		@@CACHE[self.accessor][self.class] = {} unless
			@@CACHE[self.accessor][self.class]
		@@CACHE[self.accessor][self.class][self.record_id] = self
	end

	def decache
		@@CACHE[self.accessor] = {} unless @@CACHE[self.accessor]
		@@CACHE[self.accessor][self.class] = {} unless
			@@CACHE[self.accessor][self.class]
		@@CACHE[self.accessor][self.class].delete(self.record_id)
	end

	def flush(initiator = nil)
		# We're looping back if this was started by us.  Don't do
		# that!
		return if(initiator.id == self.id)
		initiator = self if initiator == nil
		local_id = self.record_id
		if(local_id == -2)
  			$stderr.puts "ERROR:  You're working with an invalid "+
			             "object -- " + self.to_s
			return
 		elsif(@insertme)
  			@hashtable.delete('record_id') if local_id < 0
  			@hashtable.store('record_id',@accessor.insert(self)['record_id'])
		elsif(@dirty)
			@accessor.update self
		end
		@insertme=@dirty=false
		backrefs.each do |k,v|
			eval "@#{k}.each { |member|\n"+
			     "member.hashtable['#{v[0]}']=#{self.record_id.to_s}\n"+
			     "member.dirty=true\n"+
			     "member.flush(initiator)\n"+
			     "} if @#{k}"
		end
		refs.each do |k,v|
			eval "@#{k}.flush(initiator);@hashtable['#{v[0]}'] = @#{k}.record_id.to_s" if(eval "@#{k}")
		end
		self.cache if self.class.cached
		return self
	end

	def fetchRef(refname)
  		sym = refname.type==String ? refname.intern : refname
  		ref = refs[sym]
		str =<<-EOS
@#{sym} = @hashtable['#{ref[0]}'] ?
	#{ref[1]}.new(@accessor,@hashtable['#{ref[0]}'].to_i) :
	nil
EOS
		return eval str
	end


	# Similar to fetchAll() we never know if our cache is a partial or
	# complete set, so we just fetch them all again...
	def fetchBackrefs(refname)
  		sym = refname.type==String ? refname.intern : refname
  		bref = backrefs[sym]
		rows=@accessor.listByType(bref[1],
                                  "WHERE #{bref[0]+'='+record_id.to_s}")
		str = <<-EOS
@#{sym} = DBBackedArray.new(true) { |v| v.hashtable['#{bref[0]}'] = record_id }
rows.each { |r| @#{sym} << #{bref[1]}.new(@accessor,r[primary_key_name],r)}
@#{sym}.writeable=#{bref[2]}
EOS
		eval str
		return eval "@#{sym}"
  	end

	def delete()
  		if(record_id >= 0)
			@accessor.delete self
			@record_id = -2
			@hashtable = { 'record_id' => @record_id.to_s }
			self.decache
		end
  	end

end

