#!/usr/bin/env ruby

=begin
= Synopsis

    mysql -u root -p < sample-tables.sql 
    ./sample.rb | more

= Description

This sample program demonstrates some of the capabilities
((({hashed_attr})), (({hashed_ref})), (({hashed_backref}))) of the
db-backed library.  You'll want to read through (({sample.rb})),
(({sample-classes.rb})), and (({sample-tables.sql})) to see how they
work together.

=end

$:[0,0]='../lib'

require 'dbi'
require 'db-backed'

# You should look through the files named below!
load 'sample-classes.rb'
load 'db-conf.rb'

# Connect to the database and then create a DBIAccessor object.
dbh = DBI.connect($db_path,$db_user,$db_password)
accessor = DBIAccessor.new(dbh,$db_query_strings)

######################################################
#### METHODS TO PRINT OUT DATABASE BACKED OBJECTS ####
######################################################
def print_phones(phones)
	phones.each do |phone|
		puts "\tPhone: #{phone.number} belongs to #{phone.person.name}"
	end
end

def print_people(people)
	people.each do |person|
		puts "\tPerson: #{person.name}"
		person.phones.each do |phone|
			puts "\t\tPhone Number: #{phone.number}"
		end
	end
end
######################################################
######################################################
######################################################

puts "00 - Phone is a cached class, Person is not."
puts Phone.cached
puts Person.cached

puts "01 - Counting stored people."
puts "\t"+Person.count(accessor).to_s

puts "02 - Listing current people."
print_people Person.fetchAll(accessor)

puts "03 - Creating three people."
people = []
3.times do |n|
	p = Person.new(accessor)
	# We actually flush it right away because we want to hurry up and
	# get a record_id (it doesn't get one until written out)
	p.flush
	p.last_name = 'Person'
	p.first_name = "Number #{p.record_id}"
	people << p
end

# Notice how we've stored the people in an array named people above and
# now we're just going to use that because it's faster and easier than
# reinstantiating from the database, though we could do that too if we
# wanted.
puts "04 - Creating two phones for each person."
people.each do |p|
	phones = p.phones
	2.times do |n|
		ph = Phone.new(accessor)
		ph.number = "#{p.record_id} - #{n}"
		phones << ph
	end
	p.flush
puts "\t04.5 - Testing cached attributes (this should be fast)..."
	1000.times do
		rid = p.phones[0].record_id
		Phone.new(accessor,rid)
	end
end

# Now lets make sure stuff go out to the database okay by loading each
# person by record_id from the database and comparing with what we already
# have.
puts "05 - Verifying persistent storage."
people.each do |p1|
	record_id = p1.record_id
	p2 = Person.new(accessor, record_id)
	puts "\tERR: #{p1} != #{p2}" if (p1 != p2)
end

# We'll fetch from the database, just to be cool. :)
puts "06 - Listing current people."
print_people Person.fetchAll(accessor)

# Just a different way to look at our data...
puts "07 - Listing current phones."
print_phones Phone.fetchAll(accessor)

# Make sure if foo has a phone, the phone recognizes foo as its owner.
# Because of the way hashed_backref works, this should always work.
puts "08 - Verifying cross references."
Person.fetchAll(accessor).each do |person|
	person.phones.each do |phone|
		puts "\tERR: #{person} != #{phone.person}" if
			(phone.person != person)
	end
end

puts "09 - Counting stored people."
puts "\t"+Person.count(accessor).to_s

# And delete everything we've created in this session, just so that
# subsequent runs of sample.rb will look alike.
puts "10 - Deleting new people and phones."
people.each do |p|
	p.phones.each { |ph| ph.delete }
	p.delete
end

puts "11 - Counting stored people."
puts "\t"+Person.count(accessor).to_s

# Goodbye!

dbh.disconnect
